# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging

import traits.api as t

from hyperspy.docstrings.signal import LAZYSIGNAL_DOC
from hyperspy.signal import BaseSetMetadataItems
from hyperspy.ui_registry import add_gui_method, DISPLAY_DT, TOOLKIT_DT

from .eds import EDSSpectrum, LazyEDSSpectrum
from exspy._defaults_parser import preferences


_logger = logging.getLogger(__name__)


@add_gui_method(toolkey="exspy.microscope_parameters_EDS_SEM")
class EDSSEMParametersUI(BaseSetMetadataItems):
    beam_energy = t.Float(t.Undefined, label="Beam energy (keV)")
    live_time = t.Float(t.Undefined, label="Live time (s)")
    tilt_stage = t.Float(t.Undefined, label="Stage tilt (degree)")
    azimuth_angle = t.Float(t.Undefined, label="Azimuth angle (degree)")
    elevation_angle = t.Float(t.Undefined, label="Elevation angle (degree)")
    energy_resolution_MnKa = t.Float(t.Undefined, label="Energy resolution MnKa (eV)")
    mapping = {
        "Acquisition_instrument.SEM.beam_energy": "beam_energy",
        "Acquisition_instrument.TEM.Stage.tilt_alpha": "tilt_stage",
        "Acquisition_instrument.SEM.Detector.EDS.live_time": "live_time",
        "Acquisition_instrument.SEM.Detector.EDS.azimuth_angle": "azimuth_angle",
        "Acquisition_instrument.SEM.Detector.EDS.elevation_angle": "elevation_angle",
        "Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa": "energy_resolution_MnKa",
    }


class EDSSEMSpectrum(EDSSpectrum):

    """Signal class for EDS spectra measured in an SEM."""

    _signal_type = "EDS_SEM"

    def __init__(self, *args, **kwards):
        super().__init__(*args, **kwards)
        # Attributes defaults
        if "Acquisition_instrument.SEM.Detector.EDS" not in self.metadata:
            if "Acquisition_instrument.TEM" in self.metadata:
                self.metadata.set_item(
                    "Acquisition_instrument.SEM",
                    self.metadata.Acquisition_instrument.TEM,
                )
                del self.metadata.Acquisition_instrument.TEM
        self._set_default_param()

    def get_calibration_from(self, ref, nb_pix=1):
        """Copy the calibration and all metadata of a reference.

        Primary use: To add a calibration to ripple file from INCA
        software

        Parameters
        ----------
        ref : signal
            The reference contains the calibration in its
            metadata
        nb_pix : int
            The live time (real time corrected from the "dead time")
            is divided by the number of pixel (spectrums), giving an
            average live time.

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.

        Examples
        --------
        >>> ref = exspy.data.EDS_SEM_TM002()
        >>> s = exspy.signals.EDSSEMSpectrum(ref.data)
        >>> print(s.axes_manager[0].scale)
        >>> s.get_calibration_from(ref)
        >>> print(s.axes_manager[0].scale)
        1.0
        0.01

        """

        self._original_metadata = ref.original_metadata.deepcopy()
        # Setup the axes_manager
        ax_m = self.axes_manager.signal_axes[0]
        ax_ref = ref.axes_manager.signal_axes[0]
        for _axis in [ax_m, ax_ref]:
            if not _axis.is_uniform:
                raise NotImplementedError(
                    "The function is not implemented for non-uniform axes."
                )
        ax_m.scale = ax_ref.scale
        ax_m.units = ax_ref.units
        ax_m.offset = ax_ref.offset

        # Setup metadata
        if "Acquisition_instrument.SEM" in ref.metadata:
            mp_ref = ref.metadata.Acquisition_instrument.SEM
        elif "Acquisition_instrument.TEM" in ref.metadata:
            mp_ref = ref.metadata.Acquisition_instrument.TEM
        else:
            raise ValueError(
                "The reference has no metadata.Acquisition_instrument.TEM"
                "\n nor metadata.Acquisition_instrument.SEM "
            )

        mp = self.metadata

        mp.Acquisition_instrument.SEM = mp_ref.deepcopy()

        if hasattr(mp_ref.Detector.EDS, "live_time"):
            mp.Acquisition_instrument.SEM.Detector.EDS.live_time = (
                mp_ref.Detector.EDS.live_time / nb_pix
            )

    def _load_from_TEM_param(self):
        """Transfer metadata.Acquisition_instrument.TEM to
        metadata.Acquisition_instrument.SEM

        """

        mp = self.metadata
        if mp.has_item("Acquisition_instrument.SEM") is False:
            mp.add_node("Acquisition_instrument.SEM")
        if mp.has_item("Acquisition_instrument.SEM.Detector.EDS") is False:
            mp.Acquisition_instrument.SEM.add_node("EDS")
        mp.Signal.signal_type = "EDS_SEM"

        # Transfer
        if "Acquisition_instrument.TEM" in mp:
            mp.Acquisition_instrument.SEM = mp.Acquisition_instrument.TEM
            del mp.Acquisition_instrument.TEM

    def _set_default_param(self):
        """Set to value to default (defined in preferences)"""
        mp = self.metadata
        if "Acquisition_instrument.SEM.Stage.tilt_alpha" not in mp:
            mp.set_item(
                "Acquisition_instrument.SEM.Stage.tilt_alpha",
                preferences.EDS.eds_tilt_stage,
            )
        if "Acquisition_instrument.SEM.Detector.EDS.elevation_angle" not in mp:
            mp.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.elevation_angle",
                preferences.EDS.eds_detector_elevation,
            )
        if "Acquisition_instrument.SEM.Detector.EDS.energy_resolution_MnKa" not in mp:
            mp.set_item(
                "Acquisition_instrument.SEM.Detector.EDS." "energy_resolution_MnKa",
                preferences.EDS.eds_mn_ka,
            )
        if "Acquisition_instrument.SEM.Detector.EDS.azimuth_angle" not in mp:
            mp.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.azimuth_angle",
                preferences.EDS.eds_detector_azimuth,
            )

    def set_microscope_parameters(
        self,
        beam_energy=None,
        live_time=None,
        tilt_stage=None,
        azimuth_angle=None,
        elevation_angle=None,
        energy_resolution_MnKa=None,
        display=True,
        toolkit=None,
    ):
        if set(
            [
                beam_energy,
                live_time,
                tilt_stage,
                azimuth_angle,
                elevation_angle,
                energy_resolution_MnKa,
            ]
        ) == {None}:
            tem_par = EDSSEMParametersUI(self)
            return tem_par.gui(toolkit=toolkit, display=display)
        md = self.metadata

        if beam_energy is not None:
            md.set_item("Acquisition_instrument.SEM.beam_energy", beam_energy)
        if live_time is not None:
            md.set_item("Acquisition_instrument.SEM.Detector.EDS.live_time", live_time)
        if tilt_stage is not None:
            md.set_item("Acquisition_instrument.SEM.Stage.tilt_alpha", tilt_stage)
        if azimuth_angle is not None:
            md.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.azimuth_angle", azimuth_angle
            )
        if elevation_angle is not None:
            md.set_item(
                "Acquisition_instrument.SEM.Detector.EDS.elevation_angle",
                elevation_angle,
            )
        if energy_resolution_MnKa is not None:
            md.set_item(
                "Acquisition_instrument.SEM.Detector.EDS." "energy_resolution_MnKa",
                energy_resolution_MnKa,
            )

    set_microscope_parameters.__doc__ = """
        Set the microscope parameters.

        If no arguments are given, raises an interactive mode to fill
        the values.

        Parameters
        ----------
        beam_energy: float
            The energy of the electron beam in keV
        live_time : float
            In second
        tilt_stage : float
            In degree
        azimuth_angle : float
            In degree
        elevation_angle : float
            In degree
        energy_resolution_MnKa : float
            In eV
        {}
        {}

        Examples
        --------
        >>> s = exspy.data.EDS_TEM_FePt_nanoparticles()
        >>> print('Default value %s eV' %
        >>>       s.metadata.Acquisition_instrument.
        >>>       SEM.Detector.EDS.energy_resolution_MnKa)
        >>> s.set_microscope_parameters(energy_resolution_MnKa=135.)
        >>> print('Now set to %s eV' %
        >>>       s.metadata.Acquisition_instrument.
        >>>       SEM.Detector.EDS.energy_resolution_MnKa)
        Default value 130.0 eV
        Now set to 135.0 eV

        """.format(
        DISPLAY_DT, TOOLKIT_DT
    )

    def _are_microscope_parameters_missing(self):
        """Check if the EDS parameters necessary for quantification
        are defined in metadata. If not, in interactive mode
        raises an UI item to fill the values

        """
        must_exist = (
            "Acquisition_instrument.SEM.beam_energy",
            "Acquisition_instrument.SEM.Detector.EDS.live_time",
        )

        missing_parameters = []
        for item in must_exist:
            exists = self.metadata.has_item(item)
            if exists is False:
                missing_parameters.append(item)
        if missing_parameters:
            _logger.info("Missing parameters {}".format(missing_parameters))
            return True
        else:
            return False

    def create_model(self, auto_background=True, auto_add_lines=True, *args, **kwargs):
        """Create a model for the current SEM EDS data.

        Parameters
        ----------
        auto_background : boolean, default True
            If True, adds automatically a polynomial order 6 to the model,
            using the edsmodel.add_polynomial_background method.
        auto_add_lines : boolean, default True
            If True, automatically add Gaussians for all X-rays generated in
            the energy range by an element using the edsmodel.add_family_lines
            method.
        dictionary : {None, dict}, optional
            A dictionary to be used to recreate a model. Usually generated
            using :meth:`hyperspy.model.as_dictionary`

        Returns
        -------

        model : `EDSSEMModel` instance.

        """
        from exspy.models.edssemmodel import EDSSEMModel

        model = EDSSEMModel(
            self,
            auto_background=auto_background,
            auto_add_lines=auto_add_lines,
            *args,
            **kwargs,
        )
        return model


class LazyEDSSEMSpectrum(EDSSEMSpectrum, LazyEDSSpectrum):

    """Lazy signal class for EDS spectra measured in an SEM."""

    __doc__ += LAZYSIGNAL_DOC.replace("__BASECLASS__", "EDSSEMSpectrum")



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest
from pathlib import Path

import hyperspy.api as hs
from hyperspy.components1d import Gaussian
from hyperspy.exceptions import VisibleDeprecationWarning
from exspy.signals import EELSSpectrum

my_path = Path(__file__).resolve().parent
baseline_dir = "plot_model"
default_tol = 2.0


def create_ll_signal(signal_shape=1000):
    offset = 0
    zlp_param = {"A": 10000.0, "centre": 0.0 + offset, "sigma": 15.0}
    zlp = Gaussian(**zlp_param)
    plasmon_param = {"A": 2000.0, "centre": 200.0 + offset, "sigma": 75.0}
    plasmon = Gaussian(**plasmon_param)
    axis = np.arange(signal_shape)
    data = zlp.function(axis) + plasmon.function(axis)
    ll = EELSSpectrum(data)
    ll.axes_manager[-1].offset = -offset
    ll.axes_manager[-1].scale = 0.1
    return ll


A_value_gaussian = [1000.0, 600.0, 2000.0]
centre_value_gaussian = [50.0, 20.0, 60.0]
sigma_value_gaussian = [5.0, 3.0, 1.0]
scale = 0.1


def create_sum_of_gaussians(convolved=False):
    param1 = {
        "A": A_value_gaussian[0],
        "centre": centre_value_gaussian[0] / scale,
        "sigma": sigma_value_gaussian[0] / scale,
    }
    gs1 = Gaussian(**param1)
    param2 = {
        "A": A_value_gaussian[1],
        "centre": centre_value_gaussian[1] / scale,
        "sigma": sigma_value_gaussian[1] / scale,
    }
    gs2 = Gaussian(**param2)
    param3 = {
        "A": A_value_gaussian[2],
        "centre": centre_value_gaussian[2] / scale,
        "sigma": sigma_value_gaussian[2] / scale,
    }
    gs3 = Gaussian(**param3)

    axis = np.arange(1000)
    data = gs1.function(axis) + gs2.function(axis) + gs3.function(axis)

    if convolved:
        to_convolved = create_ll_signal(data.shape[0]).data
        data = np.convolve(data, to_convolved) / sum(to_convolved)

    s = EELSSpectrum(data[:1000])
    s.axes_manager[-1].scale = scale
    return s


@pytest.mark.parametrize("binned", [True, False])
@pytest.mark.parametrize("plot_component", [True, False])
@pytest.mark.parametrize("convolved", [True, False])
@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir, tolerance=default_tol)
def test_plot_gaussian_EELSSpectrum(convolved, plot_component, binned):
    s = create_sum_of_gaussians(convolved)
    s.axes_manager[-1].is_binned == binned
    s.metadata.General.title = "Convolved: {}, plot_component: {}, binned: {}".format(
        convolved, plot_component, binned
    )

    s.axes_manager[-1].is_binned = binned
    m = s.create_model(auto_add_edges=False, auto_background=False)
    m.low_loss = create_ll_signal(1000) if convolved else None

    m.extend([Gaussian(), Gaussian(), Gaussian()])

    def set_gaussian(gaussian, centre, sigma):
        gaussian.centre.value = centre
        gaussian.centre.free = False
        gaussian.sigma.value = sigma
        gaussian.sigma.free = False

    for gaussian, centre, sigma in zip(m, centre_value_gaussian, sigma_value_gaussian):
        set_gaussian(gaussian, centre, sigma)

    m.fit()
    m.plot(plot_components=plot_component)

    def A_value(s, component, binned):
        if binned:
            return component.A.value * scale
        else:
            return component.A.value

    if convolved:
        np.testing.assert_almost_equal(A_value(s, m[0], binned), 0.014034, decimal=5)
        np.testing.assert_almost_equal(A_value(s, m[1], binned), 0.008420, decimal=5)
        np.testing.assert_almost_equal(A_value(s, m[2], binned), 0.028068, decimal=5)
    else:
        np.testing.assert_almost_equal(A_value(s, m[0], binned), 100.0, decimal=5)
        np.testing.assert_almost_equal(A_value(s, m[1], binned), 60.0, decimal=5)
        np.testing.assert_almost_equal(A_value(s, m[2], binned), 200.0, decimal=5)

    return m._plot.signal_plot.figure


@pytest.mark.parametrize(("convolved"), [False, True])
@pytest.mark.mpl_image_compare(baseline_dir=baseline_dir, tolerance=default_tol)
def test_fit_EELS_convolved(convolved):
    # Keep this test here to avoid having to add image comparison in exspy
    pytest.importorskip("exspy", reason="exspy not installed.")
    dname = my_path.joinpath("data")
    with pytest.warns(VisibleDeprecationWarning):
        cl = hs.load(dname.joinpath("Cr_L_cl.hspy"))
    cl.axes_manager[-1].is_binned = False
    cl.metadata.General.title = "Convolved: {}".format(convolved)
    ll = None
    if convolved:
        with pytest.warns(VisibleDeprecationWarning):
            ll = hs.load(dname.joinpath("Cr_L_ll.hspy"))
    m = cl.create_model(auto_background=False, low_loss=ll, GOS="hydrogenic")
    m.fit(kind="smart")
    m.plot(plot_components=True)
    return m._plot.signal_plot.figure



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest
from hyperspy.decorators import lazifyTestClass
from hyperspy._components.gaussian import Gaussian
from hyperspy._components.lorentzian import Lorentzian
import hyperspy.api as hs

import exspy
from exspy.signals import EELSSpectrum


@lazifyTestClass
class TestLinearEELSFitting:
    def setup_method(self, method):
        ll = exspy.data.EELS_low_loss(navigation_shape=())
        cl = exspy.data.EELS_MnFe(add_powerlaw=False, navigation_shape=())
        m = cl.create_model(auto_background=False)
        m[0].onset_energy.value = 637.0
        m_convolved = cl.create_model(auto_background=False, low_loss=ll)
        m_convolved[0].onset_energy.value = 637.0
        self.ll, self.cl = ll, cl
        self.m, self.m_convolved = m, m_convolved

    def test_convolved_and_std_error(self):
        m = self.m_convolved
        m.fit(optimizer="lstsq")
        linear = m.as_signal()
        std_linear = m.p_std
        m.fit(optimizer="lm")
        lm = m.as_signal()
        std_lm = m.p_std
        diff = linear - lm
        np.testing.assert_allclose(diff.data.sum(), 0.0, atol=5e-6)
        np.testing.assert_allclose(std_linear, std_lm)

    def test_nonconvolved(self):
        m = self.m
        m.fit(optimizer="lstsq")
        linear = m.as_signal()
        m.fit(optimizer="lm")
        lm = m.as_signal()
        diff = linear - lm
        np.testing.assert_allclose(diff.data.sum(), 0.0, atol=5e-6)


class TestTwinnedComponents:
    def setup_method(self):
        s = exspy.data.EDS_SEM_TM002()
        m = s.create_model()
        m2 = s.isig[5.0:15.0].create_model()
        self.m = m
        self.m2 = m2

    def test_fixed_chained_twinned_components(self):
        m = self.m
        m.fit(optimizer="lstsq")
        A = m.as_signal()

        m[4].A.free = False
        m.fit(optimizer="lstsq")
        B = m.as_signal()
        np.testing.assert_allclose(A.data, B.data, rtol=5e-5)

    def test_fit_fixed_twinned_components_and_std(self):
        m = self.m2
        m[1].A.free = False
        m.fit(optimizer="lstsq")
        lstsq_fit = m.as_signal()
        nonlinear_parameters = [p for c in m for p in c.parameters if not p._linear]
        linear_std = [para.std for para in nonlinear_parameters if para.std]

        m.fit()
        nonlinear_fit = m.as_signal()
        nonlinear_std = [para.std for para in nonlinear_parameters if para.std]

        np.testing.assert_allclose(nonlinear_fit.data, lstsq_fit.data)
        np.testing.assert_allclose(nonlinear_std, linear_std)


class TestWarningSlowMultifit:
    def setup_method(self, method):
        s = hs.data.two_gaussians().inav[0]
        s.set_signal_type("EELS")
        m = s.create_model(auto_background=False, auto_add_edges=False)
        g1 = Gaussian(centre=40)
        g2 = Gaussian(centre=55)
        m.extend([g1, g2])

        # make dummy twinning
        g2.centre.twin = g1.centre
        g2.centre.twin_function_expr = "15 + x"
        g2.A.twin = g1.A
        g2.centre.twin_function_expr = "2 * x"

        m.set_parameters_not_free(only_nonlinear=True)

        self.m = m

    def test_convolved(self):
        m = self.m
        s2 = hs.signals.Signal1D(np.ones(m.signal.data.shape))
        m.low_loss = s2
        m.convolved = True
        with pytest.warns(UserWarning, match="convolution is not supported"):
            m.multifit(optimizer="lstsq")


@pytest.mark.parametrize("multiple_free_parameters", (True, False))
@pytest.mark.parametrize("nav_dim", (0, 1, 2))
def test_expression_convolved(nav_dim, multiple_free_parameters):
    s_ref = EELSSpectrum(np.ones(100))

    # Create signal to convolve
    to_convolve_component = hs.model.components1D.Gaussian(A=100, sigma=5, centre=10)
    to_convolve = EELSSpectrum(to_convolve_component.function(np.arange(100)))
    to_convolve.axes_manager[-1].offset = -to_convolve_component.centre.value

    # Create reference signal from model with convolution
    l_ref = Lorentzian(A=100, centre=20, gamma=4)
    m_ref = s_ref.create_model(auto_add_edges=False, auto_background=False)
    m_ref.append(l_ref)
    m_ref.low_loss = to_convolve
    s = m_ref.as_signal()

    if nav_dim >= 1:
        s = hs.stack([s] * 2)
        to_convolve = hs.stack([to_convolve] * 2)
    if nav_dim == 2:
        s = hs.stack([s] * 3)
        to_convolve = hs.stack([to_convolve] * 3)

    m = s.create_model(auto_add_edges=False, auto_background=False)
    l = Lorentzian(centre=20, gamma=4)
    m.append(l)
    assert not m.convolved
    m.low_loss = to_convolve
    assert m.convolved
    m.set_parameters_not_free(only_nonlinear=True)
    with pytest.warns(UserWarning):
        m.multifit(optimizer="lstsq")

    np.testing.assert_allclose(l_ref.A.value, l.A.value)
    np.testing.assert_allclose(l_ref.centre.value, l.centre.value)
    np.testing.assert_allclose(l_ref.gamma.value, l.gamma.value)
    np.testing.assert_allclose(m.as_signal().data, s.data)
    if nav_dim in (1, 2):
        np.testing.assert_allclose(l.A.map["values"].mean(), l_ref.A.value)
        np.testing.assert_allclose(l.centre.map["values"].mean(), l_ref.centre.value)
        np.testing.assert_allclose(l.gamma.map["values"].mean(), l_ref.gamma.value)


@pytest.mark.parametrize("nav_dim", (0, 1, 2))
@pytest.mark.parametrize("convolve", (True, False))
def test_expression_multiple_linear_parameter(nav_dim, convolve):
    """
    This test checks that linear fitting works with convolution with
     - single and multidimensional fit (warning raise)
     - multiple free parameters for the same component (different code path)
    """
    s_ref = EELSSpectrum(np.ones(20))
    p_ref = hs.model.components1D.Polynomial(order=2, a0=25, a1=-50, a2=2.5)

    # Create signal to convolve
    to_convolve_component = Gaussian(A=100, sigma=5, centre=10)
    to_convolve = hs.signals.Signal1D(to_convolve_component.function(np.arange(1000)))
    to_convolve.axes_manager[-1].offset = -to_convolve_component.centre.value

    m_ref = s_ref.create_model(auto_add_edges=False, auto_background=False)
    m_ref.extend([p_ref])
    if convolve:
        m_ref.low_loss = to_convolve
    s = m_ref.as_signal()

    if nav_dim >= 1:
        s = hs.stack([s] * 2)
        if convolve:
            to_convolve = hs.stack([to_convolve] * 2)
    if nav_dim == 2:
        s = hs.stack([s] * 3)
        if convolve:
            to_convolve = hs.stack([to_convolve] * 3)

    m = s.create_model(auto_add_edges=False, auto_background=False)
    p = hs.model.components1D.Polynomial(order=2)
    m.append(p)
    assert not m.convolved
    if convolve:
        m.low_loss = to_convolve
        with pytest.warns(UserWarning):
            m.multifit(optimizer="lstsq")
    else:
        m.multifit(optimizer="lstsq")

    np.testing.assert_allclose(p_ref.a0.value, p.a0.value)
    np.testing.assert_allclose(p_ref.a1.value, p.a1.value)
    np.testing.assert_allclose(p_ref.a2.value, p.a2.value)
    np.testing.assert_allclose(m.as_signal().data, s.data)
    if nav_dim >= 1:
        np.testing.assert_allclose(p.a0.map["values"].mean(), p_ref.a0.value)
        np.testing.assert_allclose(p.a1.map["values"].mean(), p_ref.a1.value)
        np.testing.assert_allclose(p.a2.map["values"].mean(), p_ref.a2.value)


@pytest.mark.parametrize("nav_dim", (0, 1, 2))
def test_multiple_linear_parameters_convolution(nav_dim):
    s_ref = EELSSpectrum(np.ones(1000))

    # Create signal to convolve
    to_convolve_component = Gaussian(A=1000, sigma=50, centre=100)
    to_convolve = EELSSpectrum(to_convolve_component.function(np.arange(1000)))
    to_convolve.axes_manager[-1].offset = -to_convolve_component.centre.value

    l_ref1 = Lorentzian(A=100, centre=200, gamma=10)
    l_ref2 = Lorentzian(A=100, centre=600, gamma=20)

    m_ref = s_ref.create_model(auto_add_edges=False, auto_background=False)
    m_ref.extend([l_ref1, l_ref2])
    m_ref.low_loss = to_convolve
    s = m_ref.as_signal()

    if nav_dim >= 1:
        s = hs.stack([s] * 2)
        to_convolve = hs.stack([to_convolve] * 2)
    if nav_dim == 2:
        s = hs.stack([s] * 3)
        to_convolve = hs.stack([to_convolve] * 3)

    m = s.create_model(auto_add_edges=False, auto_background=False)
    l1 = Lorentzian(centre=200, gamma=10)
    l2 = Lorentzian(centre=600, gamma=20)
    m.extend([l1, l2])
    assert not m.convolved
    m.low_loss = to_convolve
    assert m.convolved
    m.set_parameters_not_free(only_nonlinear=True)
    with pytest.warns(UserWarning):
        m.multifit(optimizer="lstsq")

    np.testing.assert_allclose(l_ref1.A.value, l1.A.value)
    np.testing.assert_allclose(l_ref1.centre.value, l1.centre.value)
    np.testing.assert_allclose(l_ref1.gamma.value, l1.gamma.value)
    np.testing.assert_allclose(l_ref2.A.value, l2.A.value)
    np.testing.assert_allclose(l_ref2.centre.value, l2.centre.value)
    np.testing.assert_allclose(l_ref2.gamma.value, l2.gamma.value)
    np.testing.assert_allclose(m.as_signal().data, s.data)
    if nav_dim >= 1:
        np.testing.assert_allclose(l1.A.map["values"].mean(), l_ref1.A.value)
        np.testing.assert_allclose(l1.centre.map["values"].mean(), l_ref1.centre.value)
        np.testing.assert_allclose(l1.gamma.map["values"].mean(), l_ref1.gamma.value)
        np.testing.assert_allclose(l2.A.map["values"].mean(), l_ref2.A.value)
        np.testing.assert_allclose(l2.centre.map["values"].mean(), l_ref2.centre.value)
        np.testing.assert_allclose(l2.gamma.map["values"].mean(), l_ref2.gamma.value)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import math

from hyperspy.component import Component
from hyperspy.docstrings.parameters import FUNCTION_ND_DOCSTRING

sqrt2pi = np.sqrt(2 * np.pi)


def _calculate_shirley_background(values):
    cf = np.cumsum(values, axis=-1)
    # necessary to work with `function_nd`
    return cf[..., -1][np.newaxis].T - cf


class PESCoreLineShape(Component):

    """ """

    def __init__(self, A=1.0, FWHM=1.0, origin=0.0, ab=0.0, shirley=0.0):
        Component.__init__(self, ["A", "FWHM", "origin", "ab", "shirley"])
        self.ab.value = 0
        self.ab.free = False
        self.A.value = A
        self.FWHM.value = FWHM
        self.origin.value = origin
        self._position = self.origin

        # Boundaries
        self.A.bmin = 0.0
        self.A.bmax = None
        self.FWHM.bmin = None
        self.FWHM.bmax = None

        self.isbackground = False
        self.convolved = True

        # Gradients
        self.A.grad = self.grad_A
        self.FWHM.grad = self.grad_FWHM
        self.origin.grad = self.grad_origin
        self.ab.grad = self.grad_ab

        # Options
        self.Shirley = False
        self._whitelist["Shirley"] = None

    @property
    def Shirley(self):
        return self._Shirley

    @Shirley.setter
    def Shirley(self, value):
        self._Shirley = value
        self.shirley.free = value

    def _function(self, x, A, origin, FWHM, ab, shirley):
        """
        Given an one dimensional array x containing the energies at which
        you want to evaluate the background model, returns the background
        model for the current parameters.
        """
        f = A * np.exp(-1 * math.log(2) * ((x - (origin - ab)) / FWHM) ** 2)
        if self.Shirley:
            return _calculate_shirley_background(f) * shirley + f
        else:
            return f

    def function(self, x):
        return self._function(
            x,
            self.A.value,
            self.origin.value,
            self.FWHM.value,
            self.ab.value,
            self.shirley.value,
        )

    def function_nd(self, axis):
        """%s"""
        if self._is_navigation_multidimensional:
            x = axis[np.newaxis, :]
            A = self.A.map["values"][..., np.newaxis]
            origin = self.origin.map["values"][..., np.newaxis]
            FWHM = self.FWHM.map["values"][..., np.newaxis]
            ab = self.ab.map["values"][..., np.newaxis]
            shirley = self.shirley.map["values"][..., np.newaxis]
            return self._function(x, A, origin, FWHM, ab, shirley)
        else:
            return self.function(axis)

    function_nd.__doc__ %= FUNCTION_ND_DOCSTRING

    def grad_A(self, x):
        return self.function(x) / self.A.value

    def grad_FWHM(self, x):
        a0 = self.A.value
        a1 = self.origin.value
        a2 = self.FWHM.value
        a3 = self.ab.value
        return (
            2
            * math.log(2)
            * a0
            * (x + a3 - a1) ** 2
            * np.exp(-(math.log(2) * (x + a3 - a1) ** 2) / a2**2)
        ) / a2**3

    def grad_origin(self, x):
        a0 = self.A.value
        a1 = self.origin.value
        a2 = self.FWHM.value
        a3 = self.ab.value
        return (
            2
            * math.log(2)
            * a0
            * (x + a3 - a1)
            * np.exp(-(math.log(2) * (x + a3 - a1) ** 2) / a2**2)
        ) / a2**2

    def grad_ab(self, x):
        return -self.grad_origin(x)

    def grad_shirley(self, x):
        a0 = self.A.value
        a1 = self.origin.value
        a2 = self.FWHM.value
        a3 = self.ab.value
        f = a0 * np.exp(-1 * math.log(2) * ((x - (a1 - a3)) / a2) ** 2)
        return _calculate_shirley_background(f)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np

import hyperspy.api as hs
from hyperspy.misc.test_utils import ignore_warning

from exspy.components import SEE


def test_see():
    see = SEE(A=10, Phi=1.5, B=0.5)
    x = np.linspace(-5, 15, 10)
    np.testing.assert_allclose(
        see.function(x),
        np.array(
            [
                0.0,
                0.0,
                0.0,
                8.4375,
                0.342983001,
                0.0675685032,
                0.0236279967,
                0.010861538,
                0.005860978,
                0.003514161,
            ]
        ),
    )
    np.testing.assert_allclose(see.function(x), see.function_nd(x))

    _ = SEE(A=10, Phi=1.5, B=0.5, sigma=0)


def test_see_fit():
    # Component parameter values
    A = 10
    Phi = 2.5
    B = 0.5

    offset, scale, size = 0, 0.1, 100
    x = np.linspace(offset, scale * size, size)
    s = hs.signals.Signal1D(SEE(A=A, Phi=Phi, B=B).function(x))
    axis = s.axes_manager[0]
    axis.offset, axis.scale = offset, scale
    s.add_gaussian_noise(0.1, random_state=1)
    m = s.create_model()
    see = SEE(A=1, Phi=1.5, B=0.5)
    m.append(see)
    with ignore_warning(message="divide by zero", category=RuntimeWarning):
        m.fit(grad="analytical")
    np.testing.assert_allclose(see.A.value, A, rtol=0.1)
    np.testing.assert_allclose(see.Phi.value, Phi, rtol=0.1)
    np.testing.assert_allclose(see.B.value, B, rtol=0.1)


def test_see_function_nd():
    A, Phi, B = 10, 1.5, 0.5
    see = SEE(A=A, Phi=Phi, B=B)
    x = np.linspace(-5, 15, 10)
    s = hs.signals.Signal1D(np.array([x] * 2))

    # Manually set to test function_nd
    see._axes_manager = s.axes_manager
    see._create_arrays()
    see.A.map["values"] = [A] * 2
    see.Phi.map["values"] = [Phi] * 2
    see.B.map["values"] = [B] * 2

    values = see.function_nd(x)
    assert values.shape == (2, 10)
    for v in values:
        np.testing.assert_allclose(v, see.function(x))



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import pytest

import exspy


def test_eds_sem():
    s = exspy.data.EDS_SEM_TM002()
    assert isinstance(s, exspy.signals.EDSSEMSpectrum)
    assert s.axes_manager.navigation_dimension == 0
    assert s.metadata.Sample.elements == ["Al", "C", "Cu", "Mn", "Zr"]


def test_eds_tem():
    s = exspy.data.EDS_TEM_FePt_nanoparticles()
    assert isinstance(s, exspy.signals.EDSTEMSpectrum)
    assert s.axes_manager.navigation_dimension == 0
    assert s.metadata.Sample.elements == ["Fe", "Pt"]


@pytest.mark.parametrize("navigation_shape", [(), (2,), (3, 4), (5, 6, 7)])
@pytest.mark.parametrize(
    ["add_noise", "random_state"], [[True, 0], [True, None], [False, None]]
)
def test_EELS_low_loss(add_noise, random_state, navigation_shape):
    s = exspy.data.EELS_low_loss(add_noise, random_state, navigation_shape)
    assert s.axes_manager.navigation_shape == navigation_shape


@pytest.mark.parametrize("add_powerlaw", [True, False])
@pytest.mark.parametrize("navigation_shape", [(1,), (2,)])
@pytest.mark.parametrize(
    ["add_noise", "random_state"], [[True, 0], [True, None], [False, None]]
)
def test_EELS_MnFe(add_powerlaw, add_noise, random_state, navigation_shape):
    s = exspy.data.EELS_MnFe(add_powerlaw, add_noise, random_state, navigation_shape)
    if navigation_shape == (1,):
        navigation_shape = ()
    assert s.axes_manager.navigation_shape == navigation_shape



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import functools
from pathlib import Path
import warnings

import numpy as np
from scipy import interpolate

import hyperspy.api as hs
from hyperspy.misc.math_tools import check_random_state

import exspy
from exspy.misc.eels.eelsdb import eelsdb


__all__ = [
    "EDS_SEM_TM002",
    "EDS_TEM_FePt_nanoparticles",
    "eelsdb",
    "EELS_low_loss",
    "EELS_MnFe",
]


def __dir__():
    return sorted(__all__)


_ADD_NOISE_DOCSTRING = """add_noise : bool
        If True, add noise to the signal. See note to seed the noise to
        generate reproducible noise.
    random_state : None or int or RandomState instance, default None
        Random seed used to generate the data.
    """


_RETURNS_DOCSTRING = """Returns
    -------
    :py:class:`~hyperspy._signals.eels.EELSSpectrum`
    """


def _resolve_dir():
    """Returns the absolute path to this file's directory."""
    return Path(__file__).resolve().parent


def EDS_SEM_TM002():
    """
    Load an EDS-SEM spectrum of a EDS-TM002 standard supplied by the
    Bundesanstalt für Materialforschung und -prüfung (BAM).
    The sample consists of an approximately 6 µm thick layer containing
    the elements C, Al, Mn, Cu and Zr on a silicon substrate.

    Notes
    -----
    - Sample: EDS-TM002 provided by BAM (www.webshop.bam.de)
    - SEM Microscope: Nvision40 Carl Zeiss
    - EDS Detector: X-max 80 from Oxford Instrument
    """
    file_path = _resolve_dir().joinpath("EDS_SEM_TM002.hspy")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        # load "read-only" to ensure data access regardless of install location
        return hs.load(file_path, mode="r", reader="hspy")


def EDS_TEM_FePt_nanoparticles():
    """
    Load an EDS-TEM spectrum of FePt bimetallic nanoparticles.

    Notes
    -----
    - TEM Microscope: Tecnai Osiris 200 kV D658 AnalyticalTwin
    - EDS Detector: Super-X 4 detectors Brucker
    """
    file_path = _resolve_dir().joinpath("EDS_TEM_FePt_nanoparticles.hspy")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        return hs.load(file_path, mode="r", reader="hspy")


def EELS_low_loss(add_noise=True, random_state=None, navigation_shape=(10,)):
    """
    Get an artificial low loss electron energy loss spectrum.

    The zero loss peak is offset by 4.1 eV.

    Parameters
    ----------
    %s
    navigation_shape : tuple
        The shape of the navigation space.

    %s

    Example
    -------
    >>> s = exspy.data.low_loss(navigation_shape=(10, 10))
    >>> s.plot()

    See also
    --------
    EELS_MnFe

    """
    random_state = check_random_state(random_state)

    x = np.arange(-10, 40, 0.2)
    zero_loss = hs.model.components1D.Gaussian(A=500, centre=0, sigma=0.8)
    plasmon = hs.model.components1D.Gaussian(A=100, centre=15, sigma=20)

    if navigation_shape:
        s = hs.signals.Signal1D(np.ones(navigation_shape[::-1] + x.shape))
        plasmon._axes_manager = s.axes_manager
        plasmon._create_arrays()
        plasmon.centre.map["values"][:] = random_state.uniform(
            low=14.5, high=15.5, size=navigation_shape[::-1]
        )
        plasmon.centre.map["is_set"][:] = True
        plasmon.A.map["values"][:] = random_state.uniform(
            low=50, high=70, size=navigation_shape[::-1]
        )
        plasmon.A.map["is_set"][:] = True
        plasmon.sigma.map["values"][:] = random_state.uniform(
            low=1.8, high=2.2, size=navigation_shape[::-1]
        )
        plasmon.sigma.map["is_set"][:] = True

    data = np.broadcast_to(zero_loss.function(x), navigation_shape[::-1] + x.shape)
    data = data + plasmon.function_nd(x)

    if add_noise:
        data = data + random_state.uniform(size=len(x))

    from exspy.signals import EELSSpectrum

    s = EELSSpectrum(data)
    s.axes_manager[-1].offset = x[0]
    s.axes_manager[-1].scale = x[1] - x[0]
    s.metadata.General.title = "Artifical low loss EEL spectrum"
    s.axes_manager[-1].name = "Electron energy loss"
    s.axes_manager[-1].units = "eV"
    s.set_microscope_parameters(
        beam_energy=200, convergence_angle=26, collection_angle=20
    )

    return s


EELS_low_loss.__doc__ %= (_ADD_NOISE_DOCSTRING, _RETURNS_DOCSTRING)


def EELS_MnFe(
    add_powerlaw=True, add_noise=True, random_state=None, navigation_shape=(10,)
):
    """
    Get an artificial core loss electron energy loss spectrum.

    Similar to a Mn-L32 edge from a perovskite oxide.

    Some random noise is also added to the spectrum, to simulate
    experimental noise.

    Parameters
    ----------
    add_powerlaw : bool
        If True, adds a powerlaw background to the spectrum.
        Default is False.
    %s
    navigation_shape : tuple
        The shape of the navigation space. Must be of length 1.

    %s

    Example
    -------
    >>> import exspy
    >>> s = exspy.data.EELS_Mn()
    >>> s.plot()

    With the powerlaw background

    >>> s = exspy.data.EELS_Mn(add_powerlaw=True)
    >>> s.plot()

    See also
    --------
    EELS_low_loss

    """
    if len(navigation_shape) > 1:
        raise ValueError("`navigation_shape` must be of length 1.")

    random_state = check_random_state(random_state)

    x = np.arange(400, 800, 0.25)
    arctan_Mn = exspy.components.EELSArctan(A=1, k=0.2, x0=640)
    arctan_Fe = exspy.components.EELSArctan(A=1, k=0.2, x0=708)
    Mn_l3 = hs.model.components1D.Gaussian(A=150, centre=640, sigma=4)
    Mn_l2 = hs.model.components1D.Gaussian(A=75, centre=655, sigma=4)
    Fe_l3 = hs.model.components1D.Gaussian(A=150, centre=708, sigma=4)
    Fe_l2 = hs.model.components1D.Gaussian(A=50, centre=730, sigma=3)

    if len(navigation_shape) == 0 or navigation_shape == (1,):
        Mn = 0.5
        Fe = 0.5
    else:
        Mn = np.array([1, 1, 0.75, 0.5, 0])
        Fe = np.array([0, 0, 0.25, 0.5, 1])
        Mn_interpolate = interpolate.interp1d(np.arange(0, len(Mn)), Mn)
        Fe_interpolate = interpolate.interp1d(np.arange(0, len(Fe)), Fe)
        Mn = Mn_interpolate(np.linspace(0, len(Mn) - 1, navigation_shape[0]))
        Fe = Fe_interpolate(np.linspace(0, len(Fe) - 1, navigation_shape[0]))

    def get_data(component, element_distribution):
        data_ = np.broadcast_to(component.function(x), navigation_shape + x.shape)
        return (data_.T * element_distribution).T

    arctan_Mn_data = get_data(arctan_Mn, Mn)
    Mn_l3_data = get_data(Mn_l3, Mn)
    Mn_l2_data = get_data(Mn_l2, Mn)
    arctan_Fe_data = get_data(arctan_Fe, Fe)
    Fe_l3_data = get_data(Fe_l3, Fe)
    Fe_l2_data = get_data(Fe_l2, Fe)

    data = (
        arctan_Mn_data
        + Mn_l3_data
        + Mn_l2_data
        + arctan_Fe_data
        + Fe_l3_data
        + Fe_l2_data
    )

    if add_noise:
        data += random_state.uniform(size=navigation_shape + x.shape) * 0.7

    if add_powerlaw:
        powerlaw = hs.model.components1D.PowerLaw(A=10e8, r=2.9, origin=0)
        data = data + np.broadcast_to(powerlaw.function(x), navigation_shape + x.shape)

    s = exspy.signals.EELSSpectrum(data)
    s.axes_manager[-1].offset = x[0]
    s.axes_manager[-1].scale = x[1] - x[0]
    s.metadata.General.title = "Artifical core loss EEL spectrum"
    s.axes_manager[-1].name = "Electron energy loss"
    s.axes_manager[-1].units = "eV"
    s.set_microscope_parameters(
        beam_energy=200, convergence_angle=26, collection_angle=20
    )
    s.add_elements(["Fe", "Mn"])
    return s.squeeze()


EELS_MnFe.__doc__ %= (_ADD_NOISE_DOCSTRING, _RETURNS_DOCSTRING)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
from scipy import constants
from scipy.integrate import simps, cumtrapz

from hyperspy._signals.complex_signal1d import (
    ComplexSignal1D,
    LazyComplexSignal1D,
)
from hyperspy.docstrings.signal import LAZYSIGNAL_DOC
from exspy.misc.eels.tools import eels_constant


class DielectricFunction(ComplexSignal1D):

    """Signal class for dielectric functions."""

    _signal_type = "DielectricFunction"
    _alias_signal_types = ["dielectric function"]

    def get_number_of_effective_electrons(self, nat, cumulative=False):
        r"""
        Compute the number of effective electrons using the Bethe f-sum
        rule.

        The Bethe f-sum rule gives rise to two definitions of the effective
        number (see [*]_), neff1 and neff2:

        .. math::

            n_{\mathrm{eff_{1}}} = n_{\mathrm{eff}}\left(-\Im\left(\epsilon^{-1}\right)\right)

        and:

        .. math::

            n_{\mathrm{eff_{2}}} = n_{\mathrm{eff}}\left(\epsilon_{2}\right)

        This method computes and return both.

        Parameters
        ----------
        nat : float
            Number of atoms (or molecules) per unit volume of the sample.

        Returns
        -------
        neff1, neff2 : :py:class:`hyperspy.api.signals.Signal1D`
            Signal1D instances containing neff1 and neff2. The signal and
            navigation dimensions are the same as the current signal if
            `cumulative` is True, otherwise the signal dimension is 0
            and the navigation dimension is the same as the current
            signal.

        Notes
        -----
        .. [*] Ray Egerton, "Electron Energy-Loss Spectroscopy
           in the Electron Microscope", Springer-Verlag, 2011.

        """

        m0 = constants.value("electron mass")
        epsilon0 = constants.epsilon_0  # Vacuum permittivity [F/m]
        hbar = constants.hbar  # Reduced Plank constant [J·s]
        k = 2 * epsilon0 * m0 / (np.pi * nat * hbar**2)

        axis = self.axes_manager.signal_axes[0]
        if cumulative is False:
            dneff1 = k * simps(
                (-1.0 / self.data).imag * axis.axis,
                x=axis.axis,
                axis=axis.index_in_array,
            )
            dneff2 = k * simps(
                self.data.imag * axis.axis, x=axis.axis, axis=axis.index_in_array
            )
            neff1 = self._get_navigation_signal(data=dneff1)
            neff2 = self._get_navigation_signal(data=dneff2)
        else:
            neff1 = self._deepcopy_with_new_data(
                k
                * cumtrapz(
                    (-1.0 / self.data).imag * axis.axis,
                    x=axis.axis,
                    axis=axis.index_in_array,
                    initial=0,
                )
            )
            neff2 = self._deepcopy_with_new_data(
                k
                * cumtrapz(
                    self.data.imag * axis.axis,
                    x=axis.axis,
                    axis=axis.index_in_array,
                    initial=0,
                )
            )

        # Prepare return
        neff1.metadata.General.title = (
            r"$n_{\mathrm{eff}}\left(-\Im\left(\epsilon^{-1}\right)\right)$ "
            "calculated from "
            + self.metadata.General.title
            + " using the Bethe f-sum rule."
        )
        neff2.metadata.General.title = (
            r"$n_{\mathrm{eff}}\left(\epsilon_{2}\right)$ "
            "calculated from "
            + self.metadata.General.title
            + " using the Bethe f-sum rule."
        )

        return neff1, neff2

    def get_electron_energy_loss_spectrum(self, zlp, t):
        """
        Compute single-scattering electron-energy loss spectrum from
        the dielectric function.

        Parameters
        ----------
        zlp: float or :py:class:`hyperspy.api.signals.BaseSignal`
            If the ZLP is the same for all spectra, the intengral of the ZLP
            can be provided as a number. Otherwise, if the ZLP intensity is not
            the same for all spectra, it can be provided as i) a Signal
            of the same dimensions as the current signal containing the ZLP
            spectra for each location ii) a Signal of signal dimension 0
            and navigation_dimension equal to the current signal containing the
            integrated ZLP intensity.
        t: None, float or :py:class:`hyperspy.api.signals.BaseSignal`
            The sample thickness in nm. If the thickness is the same for all
            spectra it can be given by a number. Otherwise, it can be provided
            as a Signal with signal dimension 0 and navigation_dimension equal
            to the current signal.

        Returns
        -------
        :py:class:`hyperspy.api.signals.BaseSignal`
        """
        for axis in self.axes_manager.signal_axes:
            if not axis.is_uniform:
                raise NotImplementedError(
                    "The function is not implemented for non-uniform axes."
                )
        data = (
            (-1 / self.data).imag
            * eels_constant(self, zlp, t).data
            * self.axes_manager.signal_axes[0].scale
        )
        s = self._deepcopy_with_new_data(data)
        s.data = s.data.real
        s.set_signal_type("EELS")
        s.metadata.General.title = "EELS calculated from " + self.metadata.General.title
        return s


class LazyDielectricFunction(DielectricFunction, LazyComplexSignal1D):

    """Lazy signal class for dielectric functions."""

    __doc__ += LAZYSIGNAL_DOC.replace("__BASECLASS__", "DielectricFunction")



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np
import math


def _F(electron_energy):
    return (1 + electron_energy / 1022) / (1 + electron_energy / 511) ** 2


def _theta_E(density, electron_energy):
    return 5.5 * density**0.3 / (_F(electron_energy) * electron_energy)


def iMFP_Iakoubovskii(density, electron_energy):
    """Estimate electron inelastic mean free path from density

    Parameters
    ----------
    density : float
        Material density in g/cm**3
    beam_energy : float
        Electron beam energy in keV

    Notes
    -----
    For details see Equation 9 in reference [*]_.

    .. [*] Iakoubovskii, K., K. Mitsuishi, Y. Nakayama, and K. Furuya.
       ‘Thickness Measurements with Electron Energy Loss Spectroscopy’.
       Microscopy Research and Technique 71, no. 8 (2008): 626–31.
       https://onlinelibrary.wiley.com/doi/10.1002/jemt.20597

    Returns
    -------
    float
        Inelastic mean free path in nanometers
    """
    theta_C = 20  # mrad
    inv_lambda = (
        11
        * density**0.3
        / (200 * _F(electron_energy) * electron_energy)
        * np.log(theta_C**2 / _theta_E(density, electron_energy) ** 2)
    )
    return 1 / inv_lambda


def iMFP_TPP2M(electron_energy, density, M, N_v, E_g):
    """Electron inelastic mean free path using TPP-2M

    Parameters
    ----------
    electron_energy : float
        Electron beam energy in keV
    density : float
        Material density in g/cm**3
    M : float
        Molar mass in g / mol
    N_v : int
        Number of valence electron
    E_g : float
        Band gap in eV

    Returns
    -------
    float
        Inelastic mean free path in nanometers

    Notes
    -----
    For details see reference [*]_.

    .. [*] Shinotsuka, H., S. Tanuma, C. J. Powell, and D. R. Penn. ‘Calculations
       of Electron Inelastic Mean Free Paths. X. Data for 41 Elemental Solids over
       the 50 EV to 200 KeV Range with the Relativistic Full Penn Algorithm:
       Calculations of Electron Inelastic Mean Free Paths. X’. Surface and
       Interface Analysis 47, no. 9 (September 2015): 871–88.
       https://onlinelibrary.wiley.com/doi/10.1002/sia.5789
    """
    E = electron_energy * 1e3
    rho = density
    alpha = (1 + E / 1021999.8) / (1 + E / 510998.9) ** 2
    E_p = 28.816 * math.sqrt(N_v * rho / M)
    gamma = 0.191 / math.sqrt(rho)
    U = (E_p / 28.816) ** 2
    C = 19.7 - 9.1 * U
    D = 534 - 208 * U
    beta = -1 + 9.44 / math.sqrt(E_p**2 + E_g**2) + 0.69 * rho**0.1
    iMFP = (
        alpha
        * E
        / (E_p**2 * (beta * math.log(gamma * alpha * E) - C / E + D / E**2))
    )
    return iMFP


def iMFP_angular_correction(density, beam_energy, alpha, beta):
    """Estimate the effect of limited collection angle on EELS mean free path

    Parameters
    ----------
    density : float
        Material density in g/cm**3
    beam_energy : float
        Electron beam energy in keV
    alpha, beta : float
        Convergence and collection angles in mrad.

    Notes
    -----
    For details see Equation 9 in reference [*]_.

    .. [*] Iakoubovskii, K., K. Mitsuishi, Y. Nakayama, and K. Furuya.
       ‘Thickness Measurements with Electron Energy Loss Spectroscopy’.
       Microscopy Research and Technique 71, no. 8 (2008): 626–31.
       https://onlinelibrary.wiley.com/doi/10.1002/jemt.20597
    """
    theta_C = 20  # mrad
    A = (
        alpha**2
        + beta**2
        + 2 * _theta_E(density, beam_energy) ** 2
        + abs(alpha**2 - beta**2)
    )
    B = alpha**2 + beta**2 + 2 * theta_C**2 + abs(alpha**2 - beta**2)
    return np.log(theta_C**2 / _theta_E(density, beam_energy) ** 2) / np.log(
        A * theta_C**2 / B / _theta_E(density, beam_energy) ** 2
    )



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


from exspy.misc.eds.utils import (
    edx_cross_section_to_zeta,
    electron_range,
    get_xray_lines_near_energy,
    take_off_angle,
    xray_range,
    zeta_to_edx_cross_section,
)


__all__ = [
    "edx_cross_section_to_zeta",
    "electron_range",
    "get_xray_lines_near_energy",
    "take_off_angle",
    "xray_range",
    "zeta_to_edx_cross_section",
]


def __dir__():
    return sorted(__all__)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


"""This module provides tools to interact with The EELS Database."""
import requests
import logging

from hyperspy.defaults_parser import preferences
from hyperspy.docstrings.signal import SHOW_PROGRESSBAR_ARG
from hyperspy.external.progressbar import progressbar
from hyperspy.io import dict2signal
from rsciio.msa import parse_msa_string


_logger = logging.getLogger(__name__)


def eelsdb(
    spectrum_type=None,
    title=None,
    author=None,
    element=None,
    formula=None,
    edge=None,
    min_energy=None,
    max_energy=None,
    resolution=None,
    min_energy_compare="gt",
    max_energy_compare="lt",
    resolution_compare="lt",
    max_n=-1,
    monochromated=None,
    order=None,
    order_direction="ASC",
    verify_certificate=True,
    show_progressbar=None,
):
    r"""Download spectra from the EELS Data Base.

    Parameters
    ----------
    spectrum_type: {'coreloss', 'lowloss', 'zeroloss', 'xrayabs'}, optional
    title: string
        Search spectra titles for a text string.
    author: string, optional
        Search authors for a text string.
    element: string or list of strings, optional
        Filter for the presence of one or more element. Each string must
        correspond with a valid element symbol.
    formula: string
        Chemical formula of the sample.
    edge: {'K', 'L1', 'L2,3', 'M2,3', 'M4,5', 'N2,3', 'N4,5' 'O2,3', 'O4,5'}, optional
        Filter for spectra with a specific class of edge.
    min_energy, max_energy: float, optional
        Minimum and maximum energy in eV.
    resolution: float, optional
        Energy resolution in eV.
    resolution_compare: {"lt", "eq", "gt"}, optional, default "lt"
        "lt" to search for all spectra with resolution less than `resolution`.
        "eq" for equal, "gt" for greater than.
    min_energy_compare, max_energy_compare: {"lt", "eq", "gt"}, optional
        "lt" to search for all spectra with min/max energy less than
        `min_energy`\`max_energy`. "eq" for equal, "gt" for greater than.
        Deafault values are "gt"/"lt" for `min_energy`\`max_energy`
        respectively.
    monochromated: bool or None (default)
    max_n: int, default -1
        Maximum number of spectra to return. -1 to return all.
    order: string
        Key to sort results by. Valid keys are:
        * "spectrumType",
        * "spectrumMin",
        * "spectrumMax",
        * "stepSize",
        * "spectrumFormula",
        * "spectrumElement",
        * "spectrumUpload",
        * "source_purity",
        * "spectrumEdges",
        * "microscope",
        * "guntype",
        * "beamenergy",
        * "resolution",
        * "monochromated",
        * "acquisition_mode",
        * "convergence",
        * "collection",
        * "probesize",
        * "beamcurrent",
        * "integratetime",
        * "readouts",
        * "detector",
        * "darkcurrent",
        * "gainvariation",
        * "calibration",
        * "zeroloss_deconv",
        * "thickness",
        * "deconv_fourier_log",
        * "deconv_fourier_ratio",
        * "deconv_stephens_deconvolution",
        * "deconv_richardson_lucy",
        * "deconv_maximum_entropy",
        * "deconv_other",
        * "assoc_spectra",
        * "ref_freetext",
        * "ref_doi",
        * "ref_url",
        * "ref_authors",
        * "ref_journal",
        * "ref_volume",
        * "ref_issue",
        * "ref_page",
        * "ref_year",
        * "ref_title",
        * "otherURLs"
    order_direction : {"ASC", "DESC"}
        Sorting `order` direction.
    verify_certificate: bool
        If True, verify the eelsdb website certificate and raise an error
        if it is invalid. If False, continue querying the database if the certificate
        is invalid. (This is a potential security risk.)
    %s



    Returns
    -------
    spectra: list
        A list containing all the spectra matching the given criteria if
        any.

    """
    # Verify arguments
    if spectrum_type is not None and spectrum_type not in {
        "coreloss",
        "lowloss",
        "zeroloss",
        "xrayabs",
    }:
        raise ValueError(
            "spectrum_type must be one of 'coreloss', 'lowloss', "
            "'zeroloss', 'xrayabs'."
        )
    valid_edges = ["K", "L1", "L2,3", "M2,3", "M4,5", "N2,3", "N4,5", "O2,3", "O4,5"]
    valid_order_keys = [
        "spectrumType",
        "spectrumMin",
        "spectrumMax",
        "stepSize",
        "spectrumFormula",
        "spectrumElement",
        "spectrumUpload",
        "source_purity",
        "spectrumEdges",
        "microscope",
        "guntype",
        "beamenergy",
        "resolution",
        "monochromated",
        "acquisition_mode",
        "convergence",
        "collection",
        "probesize",
        "beamcurrent",
        "integratetime",
        "readouts",
        "detector",
        "darkcurrent",
        "gainvariation",
        "calibration",
        "zeroloss_deconv",
        "thickness",
        "deconv_fourier_log",
        "deconv_fourier_ratio",
        "deconv_stephens_deconvolution",
        "deconv_richardson_lucy",
        "deconv_maximum_entropy",
        "deconv_other",
        "assoc_spectra",
        "ref_freetext",
        "ref_doi",
        "ref_url",
        "ref_authors",
        "ref_journal",
        "ref_volume",
        "ref_issue",
        "ref_page",
        "ref_year",
        "ref_title",
        "otherURLs",
    ]
    if edge is not None and edge not in valid_edges:
        raise ValueError("`edge` must be one of %s." % ", ".join(valid_edges))

    if order is not None and order not in valid_order_keys:
        raise ValueError("`order` must be one of %s." % ", ".join(valid_order_keys))
    if order_direction is not None and order_direction not in ["ASC", "DESC"]:
        raise ValueError('`order_direction` must be "ASC" or "DESC".')
    for kwarg, label in (
        (resolution_compare, "resolution_compare"),
        (min_energy_compare, "min_energy_compare"),
        (max_energy_compare, "max_energy_compare"),
    ):
        if kwarg not in ("lt", "gt", "eq"):
            raise ValueError('`%s` must be "lt", "eq" or "gt".' % label)
    if monochromated is not None:
        monochromated = 1 if monochromated else 0
    params = {
        "type": spectrum_type,
        "title": title,
        "author": author,
        "edge": edge,
        "min_energy": min_energy,
        "max_energy": max_energy,
        "resolution": resolution,
        "resolution_compare": resolution_compare,
        "monochromated": monochromated,
        "formula": formula,
        "min_energy_compare": min_energy_compare,
        "max_energy_compare": max_energy_compare,
        "per_page": max_n,
        "order": order,
        "order_direction": order_direction,
    }

    if isinstance(element, str):
        params["element"] = element
    else:
        params["element[]"] = element

    if show_progressbar is None:
        show_progressbar = preferences.General.show_progressbar

    request = requests.get(
        "https://api.eelsdb.eu/spectra", params=params, verify=verify_certificate
    )
    spectra = []
    jsons = request.json()
    if "message" in jsons:
        # Invalid query, EELSdb raises error.
        raise IOError(
            "Please report the following error to the exSpy developers: "
            f"{jsons['message']}."
        )

    for json_spectrum in progressbar(jsons, disable=not show_progressbar):
        download_link = json_spectrum["download_link"]
        if download_link.split(".")[-1].lower() != "msa":
            _logger.exception(
                "The source file is not a msa file, please report this error "
                "to https://eelsdb.eu/about with the following details:\n"
                f"Title: {json_spectrum['title']}\nid: {json_spectrum['id']}\n"
                f"Download link: {download_link}\n"
                f"Permalink: {json_spectrum['permalink']}"
            )
            continue
        msa_string = requests.get(download_link, verify=verify_certificate).text
        try:
            s = dict2signal(parse_msa_string(msa_string)[0])
            emsa = s.original_metadata
            s._original_metadata = type(s.original_metadata)({"json": json_spectrum})
            s.original_metadata.emsa = emsa
            spectra.append(s)

        except:
            # parse_msa_string or dict2signal may fail if the EMSA file is not
            # a valid one.
            _logger.exception(
                "Failed to load the spectrum.\n"
                "Title: %s id: %s.\n"
                "Please report this error to https://eelsdb.eu/about \n"
                % (json_spectrum["title"], json_spectrum["id"])
            )

    if not spectra:
        _logger.info(
            "The EELS database does not contain any spectra matching your query"
            ". If you have some, why not submitting them "
            "https://eelsdb.eu/submit-data/ ?\n"
        )
    else:
        # Add some info from json to metadata
        # Values with units are not yet supported by HyperSpy (v0.8) so
        # we can't get map those fields.
        for s in spectra:
            if spectrum_type == "xrayabs":
                s.set_signal_type("XAS")
            json_md = s.original_metadata.json
            s.metadata.General.title = json_md.title
            if s.metadata.Signal.signal_type == "EELS":
                if json_md.get_item("elements"):
                    try:
                        # When 'No' is in the list of elements
                        # https://api.eelsdb.eu/spectra/zero-loss-c-feg-hitachi-disp-0-214-ev/
                        if json_md.elements[0].lower() != "no":
                            s.add_elements(json_md.elements)
                    except ValueError:
                        _logger.exception(
                            "The following spectrum contains invalid chemical "
                            "element information:\n"
                            "Title: %s id: %s. Elements: %s.\n"
                            "Please report this error in "
                            "https://eelsdb.eu/about \n"
                            % (json_md.title, json_md.id, json_md.elements)
                        )
                if "collection" in json_md and " mrad" in json_md.collection:
                    beta = float(json_md.collection.replace(" mrad", ""))
                    s.metadata.set_item(
                        "Acquisition_instrument.TEM.Detector.EELS.collection_angle",
                        beta,
                    )
                if "convergence" in json_md and " mrad" in json_md.convergence:
                    alpha = float(json_md.convergence.replace(" mrad", ""))
                    s.metadata.set_item(
                        "Acquisition_instrument.TEM.convergence_angle", alpha
                    )
                if "beamenergy" in json_md and " kV" in json_md.beamenergy:
                    beam_energy = float(json_md.beamenergy.replace(" kV", ""))
                    s.metadata.set_item(
                        "Acquisition_instrument.TEM.beam_energy", beam_energy
                    )
            # We don't yet support units, so we cannot map the thickness
            # s.metadata.set_item("Sample.thickness", json_md.thickness)
            s.metadata.set_item("Sample.description", json_md.description)
            s.metadata.set_item("Sample.chemical_formula", json_md.formula)
            s.metadata.set_item("General.author", json_md.author.name)
            s.metadata.set_item(
                "Acquisition_instrument.TEM.microscope", json_md.microscope
            )

    return spectra


eelsdb.__doc__ %= SHOW_PROGRESSBAR_ARG



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


from hyperspy._components.expression import Expression


class EELSArctan(Expression):
    r"""Arctan function component for EELS (with minimum at zero).

    .. math::

        f(x) = A \cdot \left( \frac{\pi}{2} +
               \arctan \left[ k \left( x-x_0 \right) \right] \right)


    ============ =============
    Variable      Parameter
    ============ =============
    :math:`A`     A
    :math:`k`     k
    :math:`x_0`   x0
    ============ =============


    Parameters
    -----------
    A : float
        Amplitude parameter. :math:`\lim_{x\to -\infty}f(x)=0` and
        :math:`\lim_{x\to\infty}f(x)=2A`
    k : float
        Slope (steepness of the step). The larger :math:`k`, the sharper the
        step.
    x0 : float
        Center parameter (:math:`f(x_0)=A`).

    """

    def __init__(self, A=1.0, k=1.0, x0=1.0, module=["numpy", "scipy"], **kwargs):
        # To be able to still read old file versions that contain this argument
        if "minimum_at_zero" in kwargs:
            del kwargs["minimum_at_zero"]
        super().__init__(
            expression="A * (pi /2 + arctan(k * (x - x0)))",
            name="EELSArctan",
            A=A,
            k=k,
            x0=x0,
            position="x0",
            module=module,
            autodoc=False,
            **kwargs,
        )

        self.isbackground = False
        self.convolved = True



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import pytest

import exspy


def test_eels():
    s = exspy.signals.EELSSpectrum(([0, 1]))
    s0 = s.deepcopy()
    s.axes_manager[0].convert_to_non_uniform_axis()
    with pytest.raises(NotImplementedError):
        s.align_zero_loss_peak()
    with pytest.raises(NotImplementedError):
        s.create_model(low_loss=s)
    with pytest.raises(NotImplementedError):
        s.fourier_log_deconvolution(0)
    with pytest.raises(NotImplementedError):
        s.fourier_ratio_deconvolution(s)
    with pytest.raises(NotImplementedError):
        s.fourier_ratio_deconvolution(s0)
    with pytest.raises(NotImplementedError):
        s0.fourier_ratio_deconvolution(s)
    with pytest.raises(NotImplementedError):
        s.richardson_lucy_deconvolution(s)
    with pytest.raises(NotImplementedError):
        s.kramers_kronig_analysis()
    m = s.create_model()
    g = exspy.components.EELSCLEdge("N_K")
    with pytest.raises(NotImplementedError):
        m.append(g)


def test_eds():
    s = exspy.signals.EDSTEMSpectrum(([0, 1]))
    s2 = exspy.signals.EDSSEMSpectrum(([0, 1]))
    s.axes_manager[0].convert_to_non_uniform_axis()
    s2.axes_manager[0].convert_to_non_uniform_axis()
    s.set_microscope_parameters(20)
    with pytest.raises(NotImplementedError):
        s.get_calibration_from(s)
    with pytest.raises(NotImplementedError):
        s2.get_calibration_from(s2)
    m = s.create_model()
    with pytest.raises(NotImplementedError):
        m.add_family_lines("Al_Ka")
    with pytest.raises(NotImplementedError):
        m._set_energy_scale("Al_Ka", [1.0])
    with pytest.raises(NotImplementedError):
        m._set_energy_offset("Al_Ka", [1.0])
